"use strict";

// support IE8
if (!Array.prototype.indexOf)
    Array.prototype.indexOf = function(x){
        for(var i = 0; i < this.length; ++i)
            if (this[i] === x)
                return i;
        return -1;
    };

// support Function.bind
if (!Function.prototype.bind)
    Function.prototype.bind = function(thisArg){
        var self = this;
        var bindArgs = Array.prototype.slice.call(arguments, 1);
        return function(){
            return self.apply(thisArg, bindArgs.concat(Array.prototype.slice.call(arguments)));
        };
    };

function RtlCons(rtl, demandedCallback){
    this.__entries = {};
    this.__rtl = rtl;
    this.__demandedCallback = demandedCallback;

    for(var name in rtl.methods){
        this[name] = this.__makeOnDemand(name);
        this[name + "Id"] = this.__makeIdOnDemand(name);
    }
}
var rtlPrototype = {
    name: function(){return "RTL$";},
    module: function(){return this.__rtl.nodejsModule;},
    generate: function(){
        var result = "var " + this.name() + " = {\n";
        var firstEntry = true;
        for (var name in this.__entries){
            if (!firstEntry)
                result += ",\n";
            else
                firstEntry = false;
            result += "    " + name + ": " + this.__entries[name].toString();
        }
        if (!firstEntry)
            result += "\n};\n";
        else
            result = undefined;
        
        return result;
    },
    __putEntry: function(name){
        if (this.__entries[name])
            return;
        
        this.__entries[name] = this.__rtl.methods[name];
        
        var dependencies = this.__rtl.dependencies[name];
        if (dependencies)
            for(var i = 0; i < dependencies.length; ++i)
                this.__putEntry(dependencies[i]);
    },
    __makeIdOnDemand: function(name){
        return function(){
            if (this.__demandedCallback)
                this.__demandedCallback();
            this.__putEntry(name);
            return this.name() + "." + name;
        };
    },
    __makeOnDemand: function(name){
        return function(){
            var result = this[name +"Id"]() + "(";
            if (arguments.length){
                result += arguments[0];
                for(var a = 1; a < arguments.length; ++a)
                    result += ", " + arguments[a];
            }
            result += ")";
            return result;
        };
    }
};

function makeRTL(rtl, demandedCallback){
    function RTL(){
        RtlCons.apply(this, arguments);
    }

    rtl.methods.extend(RTL, rtl.base);
    for(var m in rtlPrototype)
        RTL.prototype[m] = rtlPrototype[m];
    return new RTL(rtl, demandedCallback);
}

exports.makeRTL = makeRTL;
